/*----------------------------------------------------------------------
PROGRAM: ms_eigsearch3.c
DATE:    10/14/93
AUTHOR:  Baback Moghaddam, baback@media.mit.edu
------------------------------------------------------------------------


  Multiple-Scale Local feature search by eigentemplates

  This routine looks for 2 (3) BF files in the datapath (defined below):

  1. features.bf  which is an N-by-2 matrix defining the size (row,col)
     of each of the N templates.

  2. template[n].bf  where n=1...N is the eigenvector ROW matrix where
     the 1st row is the mean feature, and the remaining rows
     are the principal eigenvectors 1:M. 

     NOTE: The eigenvectors are stored in column-order (as in MATLAB)

  3. variances.bf which is an N-by-M matrix, where each row represents
     the rank-ordered eigenvalues associated with that feature


     NOTE: The set of scales over which the search is to be performed
           is provided in an ASCII file of the format:

	   s1
	   s2
	   .
	   .
	   .
	   sN

----------------------------------------------------------------------

   NOTE: In this version if unitmap'ing is selected it will use
         the Mahalanobis mean/stddev in addition to the standard
	 distance computation.

------------------------------------------------------------------------

     Configure this code using the following defines:
     
     DO_VAR:     will set each pixel whose surrond's variance
                 is < VAR_THRESHOLD to VAR_MAXVALUE

----------------------------------------------------------------------

   NOTE: Fixed the umap_musigma_score problem
	       
---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include "util.h"
#include "io.h"
#include "matrix.h"
#include "affine.h"

/* ----------- CONFIGURES ---------------- */

#define DO_VAR                  0     /*  variance check     */
#define VAR_THRESHOLD           100   /*  threshold for low-var patches */
#define VAR_MAXVALUE            1e7   /*  replacement in distance map   */

#define MAX_NUM_TEMPLATES       5
#define MAX_NUM_SCALES          32
#define MAX_NUM_EIGENVECTORS    128   /* for static storage in dffs() */
#define MAX_VECTOR_LENGTH       128    /* for static storage in mahalanobis_distance() */
#define MAX_CHARS               512

/* ----------- Command-Line Parsing Stuff ------- */

extern int optind;
extern char *optarg;
char *progname;          /* used to store the name of the program  */
char comline[170];       /* used to store the entire command line  */

#define OPTIONS "i:l:o:d:n:p:a:b:k:s:mguec"

char *usage = "\t-i indir -l list -s scalefile [-d datadir]\n\
\t\t\t[-o outfile] [-a first_ev] [-b last_ev] [-k feature]\n\
\t\t\t[-g] [-u] [-m] [-e] [-c]\n";

char *help = "\
Multiscale Eigentemplate Search\n\n\
-i indir   \t input directory\n\
-l listfile\t ASCII file listing indir files to process (one per line)\n\
-s scalefile\t ASCII file of scales\n\
-d datadir \t data dir holding eigentemplate BF files (default = ./)\n\
-o outfile \t output file (default = ./ms_eigsearch.out)\n\
-a first_ev\t first eigenvector (default = 1)\n\
-b last_ev \t last eigenvector  (default = 5)\n\
-k feature \t search for the k-th feature only\n\
-g         \t pre-process with graymap\n\
-u         \t pre-process with unitmap\n\
-m         \t use Mahalanobis distance instead of DFFS\n\
-c         \t use spatial priors\n\
-e         \t output detection maps (in indir/Maps)\n";


/* --------- Function Prototypes ---------- */

float dffs(float *patch, int N, float **eigvectors, int first, int last);
float mahalanobis(float *patch, int N, float **eigvectors, float *eigvalues,
		  int first, int last);
float graymap(float *patch, int N);
float unitmap(float *patch, int N);
void statistics(float *p, int N, float *mean, float *sigma);
float umap_musigma_distance(float mean, float sigma, float **umap_musigma);
float mahalanobis_distance(float *x, int N, float *Mu, float **inv_Sigma);


/* -------- Globals ---------------- */

float **template[MAX_NUM_TEMPLATES];
float **errormap[MAX_NUM_TEMPLATES];
float **variances;
float **umap_musigma[MAX_NUM_TEMPLATES];
float **spatial_musigma[MAX_NUM_TEMPLATES];


/* ------- Command Line Defaults ------------- */

int do_dmdumps = 0; /* default is NOT to do errormap dumps  */
int do_graymap = 0; /* default is NOT to graymap each patch */
int do_unitmap = 0; /* default is NOT to unitmap each patch */
int do_mahalanobis = 0; /* default is NOT to do this computation */
int do_spatial = 0; /* default is NOT to add [x,y] Mahalanobis dist */
int eig_first = 1;  /* search using first 5 eigenvectors by default */
int eig_last  = 5;




/*----------------------------------------------------------------------*/
/* ---------------------------- MAIN ---------------------------------- */

main(int argc, char *argv[])
{

  register int i,j,k,l,ii,jj;
  int f,s,c,feature=1,nframe,nfeatures,sets, bytes_pixel;
  int nrow, ncol, max_N, xc, yc;
  char command[MAX_CHARS],indir[MAX_CHARS],infile[MAX_CHARS];
  char listfile[MAX_CHARS], scalefile[MAX_CHARS], line[MAX_CHARS];
  char datapath[MAX_CHARS],outdir[MAX_CHARS],filename[MAX_CHARS];
  char outfile[MAX_CHARS];

  float **image, **image_orig;
  float fval1, fval2, fval, mean, sigma;
  float maxerror;
  FILE *fp, *fp2;         /* for output values dump */

  int rowl[MAX_NUM_TEMPLATES],rowr[MAX_NUM_TEMPLATES];
  int coll[MAX_NUM_TEMPLATES],colr[MAX_NUM_TEMPLATES];
  int imin[MAX_NUM_TEMPLATES],imax[MAX_NUM_TEMPLATES];
  int jmin[MAX_NUM_TEMPLATES],jmax[MAX_NUM_TEMPLATES];
  int N_dim[MAX_NUM_TEMPLATES]; /* the dimensionality of each template */
  int M_dim[MAX_NUM_TEMPLATES]; /* the # of eigenvectors for each template */

  int numscales;
  float scales[MAX_NUM_SCALES];
  int rowm[MAX_NUM_SCALES][MAX_NUM_TEMPLATES];
  int colm[MAX_NUM_SCALES][MAX_NUM_TEMPLATES];
  float error[MAX_NUM_SCALES][MAX_NUM_TEMPLATES];
  float **Warp;
  float *my_vector;
  int xdim, ydim;
  int   best_scale[MAX_NUM_TEMPLATES];
  float best_error[MAX_NUM_TEMPLATES];
  int   best_rowm[MAX_NUM_TEMPLATES];
  int   best_colm[MAX_NUM_TEMPLATES];

  
  int **templatesize;
  float *patch;



  /* required input flags */
  
  int errflag   = 0;
  int inflag    = 0;
  int scaleflag = 0;
  int listflag  = 0;
  

  /* command line defaults */

  int singlefeature = 0;  
  int firstframe = 0;
  int lastframe  = 0;
  strcpy(datapath,".");               /* default datapath directory to cwd   */


  /* setup program name and command line strings */
  
  progname = argv[0];
  for (i=0; i<argc; i++)
    strcat(comline, argv[i]),strcat(comline, " ");


  /* set up default outfile name */

  sprintf(outfile,"%s.out",progname); /* default output file is progname.out */  

  /* ----------------------  Command Line Parse ------------------------ */
  
  while ((c = getopt(argc, argv, OPTIONS)) != EOF)
    switch (c) {
      
    case 'i':
      strcpy(indir, optarg);
      inflag = 1;
      break;
      
    case 'l':
      strcpy(listfile, optarg);
      listflag = 1;
      break;

    case 'd':
      strcpy(datapath, optarg);
      break;

    case 's':
      strcpy(scalefile, optarg);
      scaleflag = 1;
      break;

    case 'k':
      feature = atoi(optarg);
      singlefeature = 1;
      break;

    case 'o':
      strcpy(outfile, optarg);
      break;

    case 'a':
      eig_first  = atoi(optarg);
      break;

    case 'b':
      eig_last = atoi(optarg);
      break;

    case 'g':
      do_graymap = 1;
      break;

    case 'u':
      do_unitmap = 1;
      break;

    case 'e':
      do_dmdumps = 1;
      break;

    case 'm':
      do_mahalanobis = 1;
      break;

    case 'c':
      do_spatial = 1;
      break;

    case '?':
      errflag = 1;
      break;
      
    }
  
  
  /* command line error check */
  
  if (errflag || !inflag || !scaleflag || !listflag) {
    fprintf(stderr,"\nUSAGE: %s %s\n%s\n", progname, usage, help);
    exit(1);
  }

  sprintf(outdir,"%s/Maps",indir);   /* default output dir is indir/Maps    */



  /* -------- Load eigentemplate data --------------- */

  sprintf(filename,"%s/features.bf",datapath);
 
  { 
    float **m;
    max_N = 0;
    m = read_BIN(filename, &nfeatures, &ncol);
    templatesize = imatrix(1, nfeatures, 1, 2);
    
    for (i=1; i<=nfeatures; i++) {
      for (j=1; j<=ncol; j++)
	templatesize[i][j] = (int) m[i][j];
      N_dim[i] = templatesize[i][1]*templatesize[i][2];
      if (N_dim[i]>max_N) 
	max_N = N_dim[i];
      free_matrix(m, 1, nfeatures, 1, ncol);
    }
  }

  
  for (i=1; i<=nfeatures; i++) {
    sprintf(filename,"%s/template%d.bf",datapath,i);
    template[i] = read_BIN(filename, &nrow, &ncol);
    M_dim[i] = nrow;
    if (ncol!=N_dim[i])
      myerror("Template sizes in BF don't match those is definition file");
  }

  if (do_mahalanobis) {
    sprintf(filename,"%s/variances.bf",datapath);
    variances = read_BIN(filename, &nrow, &ncol);
    if (nrow!=nfeatures) {
      fprintf(stderr,
	      "ERROR: variances.bf has %d rows! (must have %d)\n\n",
	      nrow, nfeatures);
      exit(1);
    }
    if (eig_last>ncol) {
      fprintf(stderr,
	      "ERROR: variances.bf has %d cols! (must have atleast %d)\n\n",
	      ncol, eig_last);
      exit(1);
    }
  }
    
  if (do_unitmap) {
    float **C, **invC;
    int m, n;
    C = matrix(1,2,1,2);
    invC = matrix(1,2,1,2);
    for (k=1; k<=nfeatures; k++) {
      sprintf(filename,"%s/umap_musigma%d.bf",datapath,k);
      umap_musigma[k] = read_BIN(filename, &m, &n );
     
      /* invert the covariance matrix portion of umap_musigma[k] */
      for (i=1; i<=2; i++)
	for (j=1; j<=2; j++)
	  C[i][j] = umap_musigma[k][i+1][j];
      matrix_inverse(C, 2, invC);
      for (i=1; i<=2; i++)
	for (j=1; j<=2; j++)
	  umap_musigma[k][i+1][j] = invC[i][j];

      /* ----------- DEBUG -----------
      for (i=1; i<=3; i++) {
	for (j=1; j<=2; j++)
	  printf("%f ",umap_musigma[k][i][j]);
	printf("\n");
      }
      printf("\n\n");
      C[1][1] = 10;
      C[1][2] = 10;
      printf("dist = %f \n", 
	     mahalanobis_distance(C[1], 2, umap_musigma[k][1], &umap_musigma[k][1]));
      printf("dist = %f \n",
		    umap_musigma_distance(10, 10, umap_musigma[k]));
      exit(1);
      --------------- DEBUG ------------- */
    }  
    free_matrix(C,1,2,1,2);
    free_matrix(invC,1,2,1,2);
  }

   
  if (do_spatial) {
    float **C, **invC;
    int m, n;
    C = matrix(1,2,1,2);
    invC = matrix(1,2,1,2);
    for (k=1; k<=nfeatures; k++) {
      sprintf(filename,"%s/spatial_musigma%d.bf",datapath,k);
      spatial_musigma[k] = read_BIN(filename, &m, &n );
     
      /* invert the covariance matrix portion of umap_musigma[k] */
      for (i=1; i<=2; i++)
	for (j=1; j<=2; j++)
	  C[i][j] = spatial_musigma[k][i+1][j];
      matrix_inverse(C, 2, invC);
      for (i=1; i<=2; i++)
	for (j=1; j<=2; j++)
	  spatial_musigma[k][i+1][j] = invC[i][j];

      /* ----------- DEBUG -----------
      for (i=1; i<=3; i++) {
	for (j=1; j<=2; j++)
	  printf("%f ",spatial_musigma[k][i][j]);
	printf("\n");
      }
      printf("\n\n");
      C[1][1] = 10;
      C[1][2] = 10;
      printf("dist = %f \n", 
	     mahalanobis_distance(C[1], 2, spatial_musigma[k][1], &spatial_musigma[k][1]));
      printf("dist = %f \n",
		    spatial_musigma_distance(10, 10, spatial_musigma[k]));
      exit(1);
      --------------- DEBUG ------------- */
    }  
    free_matrix(C,1,2,1,2);
    free_matrix(invC,1,2,1,2);
  }

  /* ---- determine left/right boundaries for errormaps ----- */
  for (i=1; i<=nfeatures; i++) {
    rowl[i] = (int) templatesize[i][1]/2.0;
    rowr[i] = templatesize[i][1] - rowl[i] - 1;
    coll[i] = (int) templatesize[i][2]/2.0;
    colr[i] = templatesize[i][2] - coll[i] - 1;
  }

  patch = vector(1, max_N);
  my_vector = vector(1, MAX_VECTOR_LENGTH);


  /* ----  read indir descriptor -------- */
  
  read_descriptor(indir, &nframe, &sets, &bytes_pixel, &ncol, &nrow);
  if (sets>1) 
    myerror("Input files must be single-set DAT files!");
  image = matrix(1, nrow, 1, ncol);
  image_orig = matrix(1, nrow, 1, ncol);


  /* --- allocate errormap tensor and set window boundaries --- */

  for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) {
    errormap[k] = matrix(1, nrow, 1, ncol);
    imin[k] = rowl[k] + 1;
    imax[k] = nrow - rowr[k];
    jmin[k] = coll[k] + 1;
    jmax[k] = ncol - colr[k];
  }
  

  /* -------- Load scalefile -------------- */

  sprintf(filename,"%s",scalefile);
  if ((fp = fopen(filename, "r")) == NULL) {
    fprintf(stderr,"ERROR Could not open scale file %s \n\n", filename);
    exit(1);
  }
  i = 1;
  while (fgets(line, MAX_CHARS, fp)) {
    if (strncmp(line, "#", 1) != 0 && strlen(line)>1) {
      sscanf(line, "%f", scales+i);
      i++;
    }
  }
  numscales = i-1;
  fclose(fp);
  Warp = matrix(1, 3, 1, 3);



  /* ----  open output data file ---- */

  sprintf(filename,"%s", outfile);
  if ((fp2 = fopen(filename, "w")) == NULL) {
    fprintf(stderr,"ERROR Could not open output file %s \n\n", filename);
    exit(1);
  }
    

  /* ------- echo command line  -------- */

  fprintf(stdout,"%s\n\n",comline);
  fprintf(fp2,"# %s\n\n",comline); 


  /* ------------------------------------------------------ */
  /* -------------- MAIN Sequence Loop -------------------- */
  /* ------------------------------------------------------ */
    

  if ((fp = fopen(listfile, "r")) == NULL) {
    fprintf(stderr,"ERROR: Could not open input file %s \n\n", listfile);
    exit(1);
  }

  nframe = 0;
  while (fgets(line, MAX_CHARS, fp)) {

    if (strncmp(line, "#", 1) != 0 && strlen(line)>1) {

      nframe++;
      
      /* ---- read current frame ----- */

      sscanf(line, "%s", infile);
      sprintf(filename,"%s/%s", indir, infile);
      if (bytes_pixel == 4)
	read_RAW_float(filename, image_orig, nrow, ncol);
      else
	read_RAW(filename, image_orig, nrow, ncol);
      
      
      for (s=1; s<=numscales; s++) {
	
	/* ------- initialize error map ----- */

	for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++)
	  for (i=1; i<=nrow; i++)
	    for (j=1; j<=ncol; j++)
	      errormap[k][i][j] = 0.0;
	
	/* ------ apply scaling to image_orig ------ */

	for (i=1; i<=3; i++)
	  for (j=1; j<=3; j++)
	    Warp[i][j] = 0.0;
	Warp[1][1] = Warp[2][2] = scales[s];
	Warp[3][3] = 1.0;
	affine_warp(image_orig, image, nrow, ncol, Warp, 0.0);
	xdim = (int) (scales[s]*nrow + 0.5);
	ydim = (int) (scales[s]*ncol + 0.5);
	
	
	/* ----------------- Process current frame/scale --------------- */
	
	for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) {
	  
	  imax[k] = xdim - rowr[k];
	  jmax[k] = ydim - colr[k];
	  
	  fval = FLT_MAX;
	  for (i=imin[k]; i<=imax[k]; i++)
	    for (j=jmin[k]; j<=jmax[k]; j++) {
	      
	      c = 0;
	      for (jj=j-coll[k]; jj<=j+colr[k]; jj++)
		for (ii=i-rowl[k]; ii<=i+rowr[k]; ii++) 
		  patch[++c] = image[ii][jj];
	      
	      fval2 = 0;
	      if (do_unitmap) {
		statistics(patch, N_dim[k], &mean, &sigma);
		fval2 = umap_musigma_distance(mean, sigma, umap_musigma[k]); 
	      }

	      if (do_mahalanobis) {
		fval1 = mahalanobis(patch, N_dim[k], template[k], 
				    variances[k], eig_first, eig_last);
	      }
	      else
		fval1 = dffs(patch,N_dim[k],template[k],eig_first,eig_last);

	      if (do_unitmap) 
		fval1 += fval2;

	      
	      if (do_spatial) {
		my_vector[1] = (int) (i/scales[s]+0.5);
		my_vector[2] = (int) (j/scales[s]+0.5);
		fval1 +=
		  mahalanobis_distance(my_vector, 2,
				       spatial_musigma[k][1],
				       &spatial_musigma[k][1]);
	      }
	      
	      errormap[k][i][j] = fval1;
	      if (fval1<fval) {
		fval = fval1;
		rowm[s][k] = (int) (i/scales[s]+0.5);
		colm[s][k] = (int) (j/scales[s]+0.5);
		error[s][k] = fval;
	      }
	      
	    }
	  

	  { /* ----- best minimum block ------- */
	    float n1,n2,n3,n4,n5,n6,n7,n8,n9;

	    fval = FLT_MAX;
	    for (i=imin[k]+1; i<=imax[k]-1; i++) {
	      for (j=jmin[k]+1; j<=jmax[k]-1; j++) {
		fval1 = errormap[k][i][j];
		if (fval1<fval) {
		  n1 = errormap[k][i-1][j-1];
		  n2 = errormap[k][i-1][j];
		  n3 = errormap[k][i-1][j+1];
		  n4 = errormap[k][i][j-1];
		  n5 = errormap[k][i][j];
		  n6 = errormap[k][i][j+1];
		  n7 = errormap[k][i+1][j-1];
		  n8 = errormap[k][i+1][j];
		  n9 = errormap[k][i+1][j+1];
		  if (n5<n1 & n5<n2 & n5<n3 & n5<n4 & 
		      n5<n6 & n5<n7 & n5<n8 & n5<n9) {
		    rowm[s][k] = (int) (i/scales[s]+0.5);
		    colm[s][k] = (int) (j/scales[s]+0.5);
		    error[s][k] = fval1;
		    fval = fval1;
		  }
		}
	      }
	    }
	  } /* ---- end of best minimum block ---- */


	  fprintf(stdout,"%s\t %1.3f \t %1d \t %3d %3d \t %1.6e \n",
		  infile, scales[s], k, 
		  rowm[s][k], colm[s][k], error[s][k]);
	  
	  if (do_dmdumps) {
	    sprintf(filename,"%s/%s_s%dt%d.bf", outdir, infile, s, k);
	    write_BIN(filename, errormap[k], xdim, ydim);
	  }
	  
	} /* end of feature sequence loop */
      
	
      }  /* end of scale sequence loop */

      
      for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) { 
	best_error[k] = FLT_MAX;
	for (s=1; s<=numscales; s++) 
	  if (error[s][k]<best_error[k]) {
	    best_error[k] = error[s][k];
	    best_scale[k] = s;
	    best_rowm[k] = rowm[s][k];
	    best_colm[k] = colm[s][k];
	  }
	fprintf(fp2,"%s \t %1.3f  %1d   %3d %3d   %1.6e \n",
		infile, scales[best_scale[k]], k, best_rowm[k], 
		best_colm[k], best_error[k]);
	fflush(fp2);
      }    

    } /* end of valid listfile line if clause */
  } /* end of frame sequence loop */
  

  
  
  /* --- free up allocated matrices ----- */

  if (do_unitmap) {
    for (k=1; k<=nfeatures; k++)
      free_matrix(umap_musigma[k], 1, 3, 1, 2);
  }

  free_vector(my_vector, 1, MAX_VECTOR_LENGTH);
  free_vector(patch, 1, max_N);
  free_imatrix(templatesize, 1, nfeatures, 1, 2);
  free_matrix(image, 1, nrow, 1, ncol);
  free_matrix(image_orig, 1, nrow, 1, ncol);
  free_matrix(Warp, 1, 3, 1, 3);
 
  fclose(fp);
  fclose(fp2); 
  
  return 0;

}



/*------------------------------------------------------------------- */
/* -------------------- end of main() ------------------------------- */
/*------------------------------------------------------------------- */


float dffs(float *patch, int N, float **eigvectors, int first, int last)
{
  register int i,j,k;
  float error,sum,mean,var=1.0,std,E,Ec,Em;
  int M = 1;
  
  static float C[MAX_NUM_EIGENVECTORS];


  if (first>0)
    M = last - first + 1;
  
  /* compute the first 2 moments of image patch */
 
  if (DO_VAR) {
    statistics(patch, N, &mean, &var);
    var = SQR(var);
  }


  /* graymap/unitmap the patch */

  if (do_graymap)  
    graymap(patch, N);
  else if (do_unitmap)
    unitmap(patch, N);

  /* subtract the mean from the patch (note mean is first row)
     and compute the Energy of the patch                       */

  E = 0.0;
  for (i=1; i<=N; i++) {
    patch[i] -= eigvectors[1][i];
    E += SQR(patch[i]);
  }


  /* project onto the appropriate eigenvectors */
  
  Ec = Em = 0.0;   /* if first=0 then this is just mean-template SSD */
  if (first>0) {   
    for (k = first+1; k<=last+1; k++) {
      sum = 0.0;
      for (i=1; i<=N; i++)
	sum += patch[i] * eigvectors[k][i];
      C[k] = sum;
      Ec += SQR(C[k]);
    }
  }

  error = (E - Ec);
  if (DO_VAR)
    if (var<VAR_THRESHOLD)
      error = VAR_MAXVALUE;


  return error;
}

/* ---------------------------------------------------------------------- */

float mahalanobis(float *patch, int N, float **eigvectors, float *eigvalues,
		  int first, int last)
{
  register int i,j,k;
  float error,sum,mean,var=1.0,sigma,E,Ec,Em,val;
  int M = 1;
  
  static float C[MAX_NUM_EIGENVECTORS];


  if (first>0)
    M = last - first + 1;
  
  /* compute the first 2 moments of image patch */
  
  if (DO_VAR) {
    statistics(patch, N, &mean, &sigma);
    var = SQR(sigma);
  }


  /* graymap/unitmap the patch */

  if (do_graymap)  
    graymap(patch, N);
  else if (do_unitmap)
    unitmap(patch, N);

  /* subtract the mean from the patch (note mean is first row)
     and compute the Energy of the patch                       */

  E = 0.0;
  for (i=1; i<=N; i++) {
    patch[i] -= eigvectors[1][i];
    E += SQR(patch[i]);
  }


  /* project onto the appropriate eigenvectors */
  
  Ec = Em = 0.0;   /* if first=0 then this is just mean-template SSD */
  if (first>0) {   
    for (k = first+1; k<=last+1; k++) {
      sum = 0.0;
      for (i=1; i<=N; i++)
	sum += patch[i] * eigvectors[k][i];
      C[k] = sum;
      val = SQR(C[k]);
      Ec += val;
      Em += val/eigvalues[k-1];   /* FIX: used to be [k] */
    }
  }


  if (DO_VAR)
    if (var<VAR_THRESHOLD)
      error = VAR_MAXVALUE;

  error = Em + (E - Ec)/eigvalues[last+1];  /* FIX: used to be [last+2] */
  return error;
}


/*---------------------------------------------------------------------- */


float graymap(float *p, int N)

/* Maps the N-dimensional float vector p from its min/max to 0/255 */

{

  register int i;
  float vmin,vmax,range,scale;

  vmin = FLT_MAX;
  vmax = -vmin;

  for (i=1; i<=N; i++) {
    if (p[i]<vmin)
      vmin = p[i];
    if (p[i]>vmax)
      vmax = p[i];
  }

  range = vmax-vmin;
  if (range==0.0)         /* if no variation don't modify array */
    return 0;

  scale = 255.0/range;

  for (i=1; i<=N; i++)
    p[i] = (int) (0.5 + scale*(p[i]-vmin));

  return range;
}



/*---------------------------------------------------------------------- */


float unitmap(float *p, int N)

/* Maps the N-dimensional float vector p to zero-mean/unit-std dev. */
/* It returns the standard deviation (or zero if sigma<FLT_MIN).    */

{

  register int i;
  float mean = 0, sigma = 0;

  statistics(p, N, &mean, &sigma);

  if (sigma>FLT_MIN) {
    for (i=1; i<=N; i++)
      p[i] = (p[i]-mean)/sigma;
    return sigma;
  }
  else {
    for (i=1; i<=N; i++)
      p[i] -= mean;
    return 0;
  }
}


/* ---------------------------------------------------------------------- */

void statistics(float *p, int N, float *mean, float *sigma)

/* computes the mean and standard deviation of the vector p */

{
  register int i;
  float val, mymean = 0, mysigma = 0; 

  for (i=1; i<=N; i++)
    mymean += p[i];
  mymean /= N;

  for (i=1; i<=N; i++) {
    val = p[i] - mymean;
    mysigma += val*val;
  }
  mysigma = sqrt(mysigma/(N-1));

  *mean = mymean;
  *sigma = mysigma;
}


/* ---------------------------------------------------------------------- */

float umap_musigma_distance(float mean, float sigma, float **umap_musigma)

/* This function takes the mean and sigma of the patch and returns
   the Mahalanobis distance according to the 2nd-order statistics
   in the matrix umap_musigma (whose first row is the mean, and 2nd-3rd
   the INVERSE covariance
*/

{
  register int i,j;
  float sum=0, distance=0;
  float x[3],y[3];

  x[1] = mean  - umap_musigma[1][1];
  x[2] = sigma - umap_musigma[1][2];

  for (i=1; i<=2; i++) {
    sum = 0;
    for (j=1; j<=2; j++)
      sum += umap_musigma[i+1][j]*x[j];
    y[i] = sum;
  }

  distance = 0;
  for (i=1; i<=2; i++)
    distance += x[i]*y[i];

  return distance;
}

/* ---------------------------------------------------------------------- */

float mahalanobis_distance(float *x, int N, float *Mu, float **inv_Sigma)

/*
  This function returns the Mahalanobis distance for a feature vector
*/

{
  register int i,j;
  float sum=0, distance=0;
  float x_n[MAX_VECTOR_LENGTH];

  for (i=1; i<=N; i++)
    x_n[i] = x[i] - Mu[i];

  for (i=1; i<=N; i++) {
    sum = 0;
    for (j=1; j<=N; j++)
      sum += inv_Sigma[i][j]*x_n[j];
    distance += x_n[i]*sum;
  }

  return distance;
}

